/*
 * Copyright (c) 2004, 2005 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2005 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Id: osm_sm_info_get_ctrl.h 1743 2005-02-06 09:38:35Z shaharf $
 */


/*
 * Abstract:
 * 	Declaration of osm_smi_get_ctrl_t.
 *	This object represents a controller that receives the IBA SMInfo
 *	query from a node.
 *	This object is part of the OpenSM family of objects.
 *
 * Environment:
 * 	Linux User Mode
 *
 * $Revision: 1.4 $
 */

#ifndef _OSM_SM_INFO_GET_CTRL_H_
#define _OSM_SM_INFO_GET_CTRL_H_

#ifdef __cplusplus
#  define BEGIN_C_DECLS extern "C" {
#  define END_C_DECLS   }
#else /* !__cplusplus */
#  define BEGIN_C_DECLS
#  define END_C_DECLS
#endif /* __cplusplus */

BEGIN_C_DECLS

/****h* OpenSM/SM Info Get Controller
* NAME
*	SM Info Get Controller
*
* DESCRIPTION
*	The SM Info Get Controller object encapsulates the information
*	needed to handle the SwitchInfo query from a node.
*
*	The SM Info Get Controller object is thread safe.
*
*	This object should be treated as opaque and should be
*	manipulated only through the provided functions.
*
* AUTHOR
*	Steve King, Intel
*
*********/
/****s* OpenSM: SM Info Get Controller/osm_smi_get_ctrl_t
* NAME
*	osm_smi_get_ctrl_t
*
* DESCRIPTION
*	SM Info Get Controller structure.
*
*	This object should be treated as opaque and should
*	be manipulated only through the provided functions.
*
* SYNOPSIS
*/
typedef struct _osm_smi_get_ctrl
{
	osm_sminfo_t				*p_smi;
	osm_mad_pool_t				*p_pool;
	osm_log_t					*p_log;
	cl_plock_t					*p_lock;
	cl_dispatcher_t				*p_disp;

} osm_smi_get_ctrl_t;
/*
* FIELDS
*	p_smi
*		Pointer to the SM Info object of the SM.
*
*	p_pool
*		Pointer to the MAD pool.
*
*	p_log
*		Pointer to the log object.
*
*	p_lock
*		Pointer to the serializing lock.
*
*	p_disp
*		Pointer to the Dispatcher.
*
* SEE ALSO
*	SM Info Get Controller object
*********/

/****f* OpenSM: SM Info Get Controller/osm_smi_get_ctrl_construct
* NAME
*	osm_smi_get_ctrl_construct
*
* DESCRIPTION
*	This function constructs a SM Info Get Controller object.
*
* SYNOPSIS
*/
void osm_smi_get_ctrl_construct(
	IN osm_smi_get_ctrl_t* const p_ctrl );
/*
* PARAMETERS
*	p_ctrl
*		[in] Pointer to a SM Info Get Controller object to construct.
*
* RETURN VALUE
*	This function does not return a value.
*
* NOTES
*	Allows calling osm_smi_get_ctrl_init, osm_smi_get_ctrl_destroy,
*	and osm_smi_get_ctrl_is_inited.
*
*	Calling osm_smi_get_ctrl_construct is a prerequisite to calling any other
*	method except osm_smi_get_ctrl_init.
*
* SEE ALSO
*	SM Info Get Controller object, osm_smi_get_ctrl_init,
*	osm_smi_get_ctrl_destroy, osm_smi_get_ctrl_is_inited
*********/

/****f* OpenSM: SM Info Get Controller/osm_smi_get_ctrl_destroy
* NAME
*	osm_smi_get_ctrl_destroy
*
* DESCRIPTION
*	The osm_smi_get_ctrl_destroy function destroys the object, releasing
*	all resources.
*
* SYNOPSIS
*/
void osm_smi_get_ctrl_destroy(
	IN osm_smi_get_ctrl_t* const p_ctrl );
/*
* PARAMETERS
*	p_ctrl
*		[in] Pointer to the object to destroy.
*
* RETURN VALUE
*	This function does not return a value.
*
* NOTES
*	Performs any necessary cleanup of the specified
*	SM Info Get Controller object.
*	Further operations should not be attempted on the destroyed object.
*	This function should only be called after a call to
*	osm_smi_get_ctrl_construct or osm_smi_get_ctrl_init.
*
* SEE ALSO
*	SM Info Get Controller object, osm_smi_get_ctrl_construct,
*	osm_smi_get_ctrl_init
*********/

/****f* OpenSM: SM Info Get Controller/osm_smi_get_ctrl_init
* NAME
*	osm_smi_get_ctrl_init
*
* DESCRIPTION
*	The osm_smi_get_ctrl_init function initializes a
*	SM Info Get Controller object for use.
*
* SYNOPSIS
*/
ib_api_status_t osm_smi_get_ctrl_init(
	IN osm_smi_get_ctrl_t* const p_ctrl,
	IN osm_sminfo_t* const p_smi,
	IN osm_mad_pool_t* const p_mad_pool,
	IN osm_log_t* const p_log,
	IN cl_plock_t* const p_lock,
	IN cl_dispatcher_t* const p_disp );
/*
* PARAMETERS
*	p_ctrl
*		[in] Pointer to an osm_smi_get_ctrl_t object to initialize.
*
*	p_smi
*		[in] Pointer to the SMInfo object of the SM.
*
*	p_mad_pool
*		[in] Pointer to the MAD pool.
*
*	p_log
*		[in] Pointer to the log object.
*
*	p_lock
*		[in] Pointer to the OpenSM serializing lock.
*
*	p_disp
*		[in] Pointer to the OpenSM central Dispatcher.
*
* RETURN VALUES
*	CL_SUCCESS if the SM Info Get Controller object was initialized
*	successfully.
*
* NOTES
*	Allows calling other SM Info Get Controller methods.
*
* SEE ALSO
*	SM Info Get Controller object, osm_smi_get_ctrl_construct,
*	osm_smi_get_ctrl_destroy, osm_smi_get_ctrl_is_inited
*********/

/****f* OpenSM: SM Info Get Controller/osm_smi_get_ctrl_is_inited
* NAME
*	osm_smi_get_ctrl_is_inited
*
* DESCRIPTION
*	Indicates if the object has been initialized with osm_smi_get_ctrl_init.
*
* SYNOPSIS
*/
boolean_t osm_smi_get_ctrl_is_inited(
	IN const osm_smi_get_ctrl_t* const p_ctrl );
/*
* PARAMETERS
*	p_ctrl
*		[in] Pointer to an osm_smi_get_ctrl_t object.
*
* RETURN VALUES
*	TRUE if the object was initialized successfully,
*	FALSE otherwise.
*
* NOTES
*	The osm_smi_get_ctrl_construct or osm_smi_get_ctrl_init must be
*	called before using	this function.
*
* SEE ALSO
*	SM Info Get Controller object, osm_smi_get_ctrl_construct,
*	osm_smi_get_ctrl_init
*********/

/****f* OpenSM: SM Info Get Controller/osm_smi_get_ctrl_process
* NAME
*	osm_smi_get_ctrl_process
*
* DESCRIPTION
*	Process the SwitchInfo attribute.
*
* SYNOPSIS
*/
ib_api_status_t osm_smi_get_ctrl_process(
	IN const osm_smi_get_ctrl_t* const p_ctrl,
	IN osm_madw_t* const p_madw );
/*
* PARAMETERS
*	p_ctrl
*		[in] Pointer to an osm_smi_get_ctrl_t object.
*
*	p_madw
*		[in] Pointer to the MAD Wrapper containing the MAD
*		that contains the node's SwitchInfo attribute.
*
* RETURN VALUES
*	CL_SUCCESS if the SwitchInfo processing was successful.
*
* NOTES
*	This function processes a SwitchInfo attribute.
*
* SEE ALSO
*	SM Info Get Controller, SM Info Response Controller
*********/
	
END_C_DECLS

#endif /* _OSM_SM_INFO_GET_CTRL_H_ */
