/*
 * Copyright (c) 2004, 2005 Voltaire, Inc. All rights reserved.
 * Copyright (c) 2002-2005 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Id: osm_subnet_config_ctrl.h 1743 2005-02-06 09:38:35Z shaharf $
 */


/*
 * Abstract:
 * 	Declaration of osm_subn_conf_ctrl_t.
 *	This object represents a controller that initiates configuration
 *	of the subnet after the discovery phase is complete.
 *	This object is part of the OpenSM family of objects.
 *
 * Environment:
 * 	Linux User Mode
 *
 * $Revision: 1.4 $
 */

#ifndef _OSM_SUBNET_CONFIG_CTRL_H_
#define _OSM_SUBNET_CONFIG_CTRL_H_

#ifdef __cplusplus
#  define BEGIN_C_DECLS extern "C" {
#  define END_C_DECLS   }
#else /* !__cplusplus */
#  define BEGIN_C_DECLS
#  define END_C_DECLS
#endif /* __cplusplus */

BEGIN_C_DECLS

/****h* OpenSM/Subnet Configuration Controller
* NAME
*	Subnet Configuration Controller
*
* DESCRIPTION
*	The Subnet Configuration Controller object encapsulates the
*	information needed to initiate a subnet configuration pass.
*	The Subnet Configuration Controller sends Dispatcher messages
*	to controllers responsible for configuring LIDs,
*	switch forwarding tables, etc.
*
*	The Subnet Configuration Controller object is thread safe.
*
*	This object should be treated as opaque and should be
*	manipulated only through the provided functions.
*
* AUTHOR
*	Steve King, Intel
*
*********/
/****s* OpenSM: Subnet Configuration Controller/osm_subn_conf_ctrl_t
* NAME
*	osm_subn_conf_ctrl_t
*
* DESCRIPTION
*	Subnet Configuration Controller structure.
*
*	This object should be treated as opaque and should
*	be manipulated only through the provided functions.
*
* SYNOPSIS
*/
typedef struct _osm_subn_conf_ctrl
{
	osm_subn_t					*p_subn;
	osm_log_t					*p_log;
	cl_plock_t					*p_lock;
	cl_dispatcher_t				*p_disp;

} osm_subn_conf_ctrl_t;
/*
* FIELDS
*	p_subn
*		Pointer to the Subnet object for this subnet.
*
*	p_log
*		Pointer to the log object.
*
*	p_lock
*		Pointer to the serializing lock.
*
*	p_disp
*		Pointer to the Dispatcher.
*
* SEE ALSO
*	Subnet Configuration Controller object
*********/

/****f* OpenSM: Subnet Configuration Controller/osm_subn_conf_ctrl_construct
* NAME
*	osm_subn_conf_ctrl_construct
*
* DESCRIPTION
*	This function constructs a Subnet Configuration Controller object.
*
* SYNOPSIS
*/
void osm_subn_conf_ctrl_construct(
	IN osm_subn_conf_ctrl_t* const p_ctrl );
/*
* PARAMETERS
*	p_ctrl
*		[in] Pointer to a Subnet Configuration Controller object to construct.
*
* RETURN VALUE
*	This function does not return a value.
*
* NOTES
*	Allows calling osm_subn_conf_ctrl_init, osm_subn_conf_ctrl_destroy,
*	and osm_subn_conf_ctrl_is_inited.
*
*	Calling osm_subn_conf_ctrl_construct is a prerequisite to calling any other
*	method except osm_subn_conf_ctrl_init.
*
* SEE ALSO
*	Subnet Configuration Controller object, osm_subn_conf_ctrl_init,
*	osm_subn_conf_ctrl_destroy, osm_subn_conf_ctrl_is_inited
*********/

/****f* OpenSM: Subnet Configuration Controller/osm_subn_conf_ctrl_destroy
* NAME
*	osm_subn_conf_ctrl_destroy
*
* DESCRIPTION
*	The osm_subn_conf_ctrl_destroy function destroys the object, releasing
*	all resources.
*
* SYNOPSIS
*/
void osm_subn_conf_ctrl_destroy(
	IN osm_subn_conf_ctrl_t* const p_ctrl );
/*
* PARAMETERS
*	p_ctrl
*		[in] Pointer to the object to destroy.
*
* RETURN VALUE
*	This function does not return a value.
*
* NOTES
*	Performs any necessary cleanup of the specified
*	Subnet Configuration Controller object.
*	Further operations should not be attempted on the destroyed object.
*	This function should only be called after a call to
*	osm_subn_conf_ctrl_construct or osm_subn_conf_ctrl_init.
*
* SEE ALSO
*	Subnet Configuration Controller object, osm_subn_conf_ctrl_construct,
*	osm_subn_conf_ctrl_init
*********/

/****f* OpenSM: Subnet Configuration Controller/osm_subn_conf_ctrl_init
* NAME
*	osm_subn_conf_ctrl_init
*
* DESCRIPTION
*	The osm_subn_conf_ctrl_init function initializes a
*	Subnet Configuration Controller object for use.
*
* SYNOPSIS
*/
ib_api_status_t osm_subn_conf_ctrl_init(
	IN osm_subn_conf_ctrl_t* const p_ctrl,
	IN osm_subn_t* const p_subn,
	IN osm_log_t* const p_log,
	IN cl_plock_t* const p_lock,
	IN cl_dispatcher_t* const p_disp );
/*
* PARAMETERS
*	p_ctrl
*		[in] Pointer to an osm_subn_conf_ctrl_t object to initialize.
*
*	p_subn
*		[in] Pointer to the Subnet object for this subnet.
*
*	p_log
*		[in] Pointer to the log object.
*
*	p_lock
*		[in] Pointer to the OpenSM serializing lock.
*
*	p_disp
*		[in] Pointer to the OpenSM central Dispatcher.
*
* RETURN VALUES
*	IB_SUCCESS if the Subnet Configuration Controller object was initialized
*	successfully.
*
* NOTES
*	Allows calling other Subnet Configuration Controller methods.
*
* SEE ALSO
*	Subnet Configuration Controller object, osm_subn_conf_ctrl_construct,
*	osm_subn_conf_ctrl_destroy, osm_subn_conf_ctrl_is_inited
*********/

/****f* OpenSM: Subnet Configuration Controller/osm_subn_conf_ctrl_is_inited
* NAME
*	osm_subn_conf_ctrl_is_inited
*
* DESCRIPTION
*	Indicates if the object has been initialized with osm_subn_conf_ctrl_init.
*
* SYNOPSIS
*/
boolean_t osm_subn_conf_ctrl_is_inited(
	IN const osm_subn_conf_ctrl_t* const p_ctrl );
/*
* PARAMETERS
*	p_ctrl
*		[in] Pointer to an osm_subn_conf_ctrl_t object.
*
* RETURN VALUES
*	TRUE if the object was initialized successfully,
*	FALSE otherwise.
*
* NOTES
*	The osm_subn_conf_ctrl_construct or osm_subn_conf_ctrl_init must be
*	called before using	this function.
*
* SEE ALSO
*	Subnet Configuration Controller object, osm_subn_conf_ctrl_construct,
*	osm_subn_conf_ctrl_init
*********/

/****f* OpenSM: Subnet Configuration Controller/osm_subn_conf_ctrl_process
* NAME
*	osm_subn_conf_ctrl_process
*
* DESCRIPTION
*	Initiate a subnet configuration pass.
*
* SYNOPSIS
*/
ib_api_status_t osm_subn_conf_ctrl_process(
	IN const osm_subn_conf_ctrl_t* const p_ctrl );
/*
* PARAMETERS
*	p_ctrl
*		[in] Pointer to an osm_subn_conf_ctrl_t object.
*
* RETURN VALUES
*	IB_SUCCESS if configuration processing was successfully
*	initiated.
*
* NOTES
*	A success status here does not indicate that
*	the entire subnet configuration process completed successfully.
*	Only one configuration pass is active at one time.
*
* SEE ALSO
*	Subnet Configuration Controller
*********/
	
END_C_DECLS

#endif /* _OSM_SUBNET_CONFIG_CTRL_H_ */
